/*
 * pf.c - powerfail daemon
 * ron@mlfarm.com,  7 Mar 1992
 * original from rac@sherpa.UUCP (Roger Cornelius), 13 May 1990
 * 
 * To install, you need a serial port with modem-control enabled.  The
 * cable from the UPS needs only two wires, the signal ground, and
 * the powerfail signal from the UPS to CD [pin 8 on a 25-pin port,
 * pin 1 on a 9-pin port, or pin-7 on the Sun IPC/IPX serial ports].
 * Redefine DEVICE and IGNORE to suit your system; IGNORE, the time in
 * seconds before a system shutdown is initiated, depends on your UPS
 * capacity.  You can also specify the device, ignoretime, or the
 * shutdown script on the command line when pf is started.  The normal
 * usage is to put pf somewhere in /etc/rc.local.
 *
 * compile:  [g]cc -O -o pf pf.c 
 */

#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <setjmp.h>
#include <syslog.h>

#define USAGE	"Usage: pf [-d device] [-i ignoretime] [-s script]\n"
#define DEVICE	"/dev/ttyb" 
#define OPEN_TM	3       
#define IGNORE	1000

jmp_buf jbuf;
char	*script[] = { "/usr/etc/shutdown", "-h", "now", "Power Failure", 0 };

main(argc,argv)
int argc;
char **argv;
{
    void sighand(), myfork();
    int didopen(), opt, ignore;
    char *device;
    extern char *optarg;

    device = DEVICE;
    ignore = IGNORE;

    if (0 != getuid())
      {
	fputs("pf: must be root\n", stderr);
	exit(1);
      }
    while ((opt = getopt(argc, argv, "d:i:s:?")) != -1) 
      {
        switch (opt) 
	  {
	  case 'd':   device = optarg; break;
	  case 'i':   ignore = atoi(optarg); break;
	  case 's':   *script = optarg; script[1] = '\0'; break;
          default :   fputs(USAGE, stderr); exit(1);
	  }
      }
    if (1 != getppid())		/* if parent not init */
      {
        myfork(); 
        setpgrp();  
        myfork(); 
      }
    close(0);			/* close files and make sure we don't */
    close(1);			/* hold any mounted fs open */
    close(2);
    chdir("/");
				/* set up syslogging, direct to */
				/* console in case logger fails, */
				/* no wait for child process */
    openlog("UPS", LOG_CONS|LOG_NOWAIT, LOG_DAEMON);

/*
 * Here we attempt to open the port attached to the UPS.  The open
 * will block, waiting for CD, until we have a power failure.  When
 * the open succeeds, we close the port, sleep the amount of time
 * specified in `ignore', then try the open again, this time setting
 * an alarm.  If the alarm runs out before the second open succeeds,
 * we only had a temporary outage and the process starts over.  If the
 * second open succeeds, we assume it's an extended failure so we exec
 * /etc/shutdown, after a warning and one final try to open the line. 
 */

  restart:
    setjmp(jbuf);
    if (didopen(device))	/* power failure */
      {
	syslog(LOG_CRIT, "Power Line Event");
	sleep(ignore);
	signal(SIGALRM, sighand);
	alarm(OPEN_TM);
	if (didopen(device)) 
	  {
	    syslog(LOG_ALERT,
		   "Power Line Failure - System Going Down in 60 seconds");
            alarm(0);
	    sleep(60);
	    if (didopen(device))
	      {
		execvp(script[0], script);
		syslog(LOG_ALERT, "Unable to shutdown system");
	      }
	    else 
	      {
		syslog(LOG_CRIT, "Power Restored");
		goto restart;
	      }
	  }
    }
    exit(0);
}



void myfork()
{
    int stat;

    stat = fork();
    if (stat < 0) 
      syslog(LOG_ALERT,"pf: fork failed");
    else if (stat > 0)  /* parent */
      exit(0);
}


/*
 * Attempt to open the device attached to the UPS
 * Should never succeed except during power failure
 */
int didopen(device)
char *device;
{
  int fd;

  fd = open(device, O_RDONLY);
  if (fd != -1)
    {
      close(fd);
      return(1);
    }
  syslog(LOG_ALERT, "pf: bogus open on UPS line");
  exit(1);
}


void sighand(sig)
int sig;
{
  longjmp(jbuf,1);
}
